unit LZSS;

interface

uses Windows, Classes, SysUtils;

const
  TLZ_LITERAL       = 1;
  TLZ_MATCH         = 0;
  TLZ_MINMATCHLEN   = 2;
  //TLZ_WINBITS       = 10;
  TLZ_MATCHBITS     = 4;
  TLZ_WINLEN        = 4095;
  TLZ_MATCHLEN      = 255;

type
  TBitStream = class
  public
    Stream: array of UCHAR;
    Position: ULONG;
    Size: ULONG;
    TempLong: ULONG;
    BitsUsed: Integer;
  public
    //Methods
    function ReadBits(NumBits: UINT): ULONG;
    procedure WriteBits(Value: UINT; NumBits: UINT);
    procedure WriteBitsFlush();
  end;

  TLZ = Class
  public
    function Compress(const Source; var Dest; Size: ULONG): ULONG;
    function Decompress(const Source; var Dest): ULONG;
  private

    DataStreamPos: ULONG;
    CompressedStreamPos: ULONG;
    Data:array of UCHAR;
    CompressedData:array of UCHAR;

    DataSize: ULONG;
    CompressedSize: ULONG;
    CompressedLong: ULONG;
    CompressedBitsUsed: Integer;

    //Functions

    function CompressLoop(): ULONG;
    procedure FindMatches(Offset, Len: PUINT);
    procedure CompressedStreamReadBits(Value: PUINT; NumBits: UINT);
    procedure CompressedStreamWriteBits(Value : UINT; NumBits: UINT);
    procedure CompressedStreamWriteBitsFlush();
end;

var
  Flags: TBitStream;
  Pairs: TBitStream;
  Literals: TBitStream;
  Lengthes: TBitStream;
  BitStream: array[0..1] of Word;
  BitCounts: Byte;
  BitsLen: Byte;
  BitMask: array[0..15] of Word = (0, 1, 3, 7, $F, $1F, $3F, $7F, $FF, $1FF, $3FF, $7FF, $FFF, $1FFF, $3FFF, $7FFF);
  p_len : Array[0..63] of BYTE =
 ($00, $00, $00, $00, $01, $01, $02, $02,
	$02, $02, $02, $02, $03, $03, $03, $03,
	$03, $03, $03, $03, $03, $03, $03, $03,
	$04, $04, $04, $04, $04, $04, $04, $04,
	$05, $05, $05, $05, $05, $05, $05, $05,
	$06, $06, $06, $06, $06, $06, $06, $06,
	$07, $07, $07, $07, $07, $07, $07, $07,
	$08, $08, $08, $08, $08, $08, $08, $08);
implementation

procedure ElliasGammaEncode(Value: UINT);
begin
  case Value of
    1:begin
        Pairs.WriteBits(Value,1);
        Lengthes.WriteBits(1,4);
      end;
    2..3:
      begin
        Pairs.WriteBits(Value, 3);
        Lengthes.WriteBits(3, 4);
      end;
    4..7:
      begin
        Pairs.WriteBits(Value, 5);
        Lengthes.WriteBits(5, 4);
      end;
    8..15:
      begin
        Pairs.WriteBits(Value, 7);
        Lengthes.WriteBits(7, 4);
      end;
    16..31:
      begin
        Pairs.WriteBits(Value, 9);
        Lengthes.WriteBits(9, 4);
      end;
    32..63:
      begin
        Pairs.WriteBits(Value, 11);
        Lengthes.WriteBits(11, 4);
      end;
    64..127:
      begin
        Pairs.WriteBits(Value, 13);
        Lengthes.WriteBits(13, 4);
      end;
    128..255:
      begin
        Pairs.WriteBits(Value, 15);
        Lengthes.WriteBits(15, 4);
      end;
  end;
end;
function ElliasGammaDecode(Value: UINT): UINT;
var
  Mask: Word;
begin
  Mask:= $8000;
  BitsLen:= 1;
  while (Mask and Value) = 0 do
  begin
    inc(BitsLen);
    Mask:= Mask shr 1;
  end;
  BitsLen:= BitsLen * 2 - 1;
  Result:= Value shr (16 - BitsLen);
end;

procedure EncodeLength(Value: UINT);
begin
  ElliasGammaEncode(Value);
end;

function DecodeLength(Value: UINT): UINT;
begin
   Result:= ElliasGammaDecode(Value);
end;

procedure EncodePosition(Value: UINT);
begin
   case Value of
    0..3:
      begin
        Pairs.WriteBits(Value, 6);
        Lengthes.WriteBits(6, 4);
      end;
    4..7:
      begin
        inc(Value, 4);
        Pairs.WriteBits(Value, 7);
        Lengthes.WriteBits(7, 4);
      end;
    8..31:
      begin
        inc(Value, $10);
        Pairs.WriteBits(Value, 8);
        Lengthes.WriteBits(8, 4);
      end;
    32..127:
      begin
        inc(Value, $40);
        Pairs.WriteBits(Value, 9);
        Lengthes.WriteBits(9, 4);
      end;
    128..255:
      begin
        inc(Value, $100);
        Pairs.WriteBits(Value, 10);
        Lengthes.WriteBits(10, 4);
      end;
    256..511:
      begin
        inc(Value, $300);
        Pairs.WriteBits(Value, 11);
        Lengthes.WriteBits(11, 4);
      end;
    512..1023:
      begin
        inc(Value, $800);
        Pairs.WriteBits(Value, 12);
        Lengthes.WriteBits(12, 4);
      end;
    1024..2047:
      begin
        inc(Value, $1400);
        Pairs.WriteBits(Value, 13);
        Lengthes.WriteBits(13, 4);
      end;
    2048..4095:
      begin
        inc(Value, $3000);
        Pairs.WriteBits(Value, 14);
        Lengthes.WriteBits(14, 4);
      end;
   end;
end;

function DecodePosition(Value: UINT): UINT;
var
  Temp: Word;
begin
  Temp:= Value shr 10;
  BitsLen:= 6 + p_len[Temp];
  Value:= Value shr (16 - Bitslen);
  case Value of
    0..3:
      begin
        Value:= Value;
      end;
    8..11:
      begin
        dec(Value, 4);
      end;
    $18..$2F:
      begin
        dec(Value, $10);
      end;
    $60..$BF:
      begin
        dec(Value, $40);
      end;
    $180..$1FF:
      dec(Value, $100);
    $400..$4FF:
      dec(Value, $300);
    $A00..$BFF:
      dec(Value, $800);
    $1800..$1BFF:
      dec(Value, $1400);
    $3800..$3FFF:
      dec(Value, $1800);
  end;
  Result:= Value;
end;


{ TBitStream }

function TBitStream.ReadBits(NumBits: UINT): ULONG;
var
  Temp: ULONG;
begin
  TempLong:= TempLong and $FFFF;
  while NumBits > 0 do
  begin
    dec(NumBits);
    if BitsUsed = 0 then
    begin
      Temp:= Stream[Position];
      inc(Position);
      TempLong:= TempLong or (Temp shl 8);
      Temp:= Stream[Position];
      inc(Position);
      TempLong:= TempLong or Temp;
      BitsUsed:= 16;
    end;
    TempLong:= TempLong shl 1;
    dec(BitsUsed);
  end;
  Result:= TempLong shr 16;
end;

procedure TBitStream.WriteBits(Value, NumBits: UINT);
begin
  while NumBits > 0 do
  begin
    dec(NumBits);
    TempLong := TempLong shl 1;
    TempLong:= TempLong or ((Value shr NumBits) and 1);
    inc(BitsUsed);
    if(BitsUsed = 8) then
    begin
      Stream[Position]:= TempLong;
      inc(Position);
      inc(Size);
      BitsUsed:= 0;
    end;
  end;  //end while
end;

procedure TBitStream.WriteBitsFlush;
begin
  if BitsUsed > 0 then
  begin
    Stream[Position]:= TempLong shl (8 - BitsUsed);
    inc(Position);
    inc(Size);
  end;
  BitsUsed:= 0;
  TempLong:= 0;
end;

{ TLZ }

function TLZ.Compress(const Source; var Dest; Size: ULONG): ULONG;
begin
  //Set up initial values
  DataStreamPos:=        0;
  CompressedStreamPos:=  0;
  CompressedLong:=       0;
  CompressedBitsUsed:=   0;
  CompressedData:=       @Dest;
  Data:=                 @Source;
  DataSize :=            Size;
  CompressedSize:=       0;
  Flags:= TBitStream.Create;
  Literals:= TBitStream.Create;
  Pairs:= TBitStream.Create;
  Lengthes:= TBitStream.Create;
  SetLength(Flags.Stream, DataSize div 8);
  SetLength(Literals.Stream, DataSize);
  SetLength(Pairs.Stream, DataSize div 2);
  SetLength(Lengthes.Stream, DataSize div 2);
  Flags.Position:= 0;
  Literals.Position:= 0;
  Lengthes.Position:= 0;
  Pairs.Position:= 0;
  Flags.TempLong:= 0;
  Literals.TempLong:= 0;
  Lengthes.TempLong:= 0;
  Pairs.TempLong:= 0;
  Flags.BitsUsed:= 0;
  Literals.BitsUsed:= 0;
  Lengthes.BitsUsed:= 0;
  Pairs.BitsUsed:= 0;
  CompressLoop;
  //CompressedStreamWriteBits(0, 8);
  CompressedStreamWriteBitsFlush;
  Flags.Free;
  Literals.Free;
  Pairs.Free;
  Lengthes.Free;
  Result := CompressedSize;
end;

procedure TLZ.CompressedStreamReadBits(Value: PUINT; NumBits: UINT);
var
  Temp: ULONG;
begin
  CompressedLong:= CompressedLong and $FFFF;
  while NumBits > 0 do
  begin
    dec(NumBits);
    if CompressedBitsUsed = 0 then
    begin
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or (Temp shl 8);
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or Temp;
      CompressedBitsUsed:= 16;
    end;
    CompressedLong:= CompressedLong shl 1;
    dec(CompressedBitsUsed);
  end;
  Value^:= CompressedLong shr 16;
end;

procedure TLZ.CompressedStreamWriteBits(Value, NumBits: UINT);
begin
  while NumBits > 0 do
  begin
    dec(NumBits);
    CompressedLong := CompressedLong shl 1;
    CompressedLong:= CompressedLong or ((Value shr NumBits) and 1);
    inc(CompressedBitsUsed);

    if(CompressedBitsUsed = 8) then
    begin
      CompressedData[CompressedStreamPos]:= CompressedLong;
      inc(CompressedStreamPos);
      inc(CompressedSize);
      dec(CompressedBitsUsed,8);
    end;
  end;  //end while
end;

function Swap(Value: Word): Word; assembler;
asm
  xchg AH, AL
end;


procedure TLZ.CompressedStreamWriteBitsFlush;
begin
  if CompressedBitsUsed > 0 then
  begin
      CompressedData[CompressedStreamPos]:= CompressedLong shl (8 - CompressedBitsUsed);
      inc(CompressedStreamPos);
      inc(CompressedSize);
      dec(CompressedBitsUsed,8);
  end;
  CompressedLong:= 0;
end;

function TLZ.CompressLoop: ULONG;
const
  Max = 128;
var // Optimization variables
  f: array[0..Max] of UINT;
  p: array[0..Max] of UINT;
  k, m: Integer;
var
  Offset, MatchLength: UINT;
  I: UINT;
  Flag: UCHAR;
  A, Mask: Byte;
  FlagsCount: LongWord;
begin
  FlagsCount:= 0;
  while DataStreamPos < DataSize do
  begin
    FindMatches(@Offset, @MatchLength);
    if MatchLength < TLZ_MINMATCHLEN then
    begin
      Literals.WriteBits(Data[DataStreamPos], 8);
      Flags.WriteBits(TLZ_LITERAL,1);
      inc(FlagsCount);
      inc(DataStreamPos);
    end
    else
    begin
      Flags.WriteBits(TLZ_MATCH, 1);
      inc(FlagsCount);
      EncodeLength(matchlength - 1);
      EncodePosition(DataStreamPos - Offset - 1);
      inc(DataStreamPos, matchlength);
    {
      f[0]:= DataStreamPos;
      f[1]:= DataStreamPos + 1;
      m:= 2;
      while true do
      begin
        p[m-2]:= Offset;
        f[m]:= f[m-2] + MatchLength;
        if (f[m] < f[m-1]+1) then break;
        inc(m);
        if Succ(m) = Max then
          break;
        DataStreamPos:= f[m-2];
        FindMatches(@Offset, @MatchLength);
      end;
      if((m and 1) = 0) then
      begin
        Literals.WriteBits(Data[DataStreamPos], 8);
        Flags.WriteBits(TLZ_LITERAL,1);
        inc(FlagsCount);
      end;
      k:= (m+1) and 1;
      while K < m-1 do
      begin
        EncodeLength(f[k+2] - f[k] - 1);
        EncodePosition(f[k] - p[k] - 1);
        Flags.WriteBits(TLZ_MATCH,1);
        inc(FlagsCount);
        DataStreamPos:= f[k+2];
        inc(k,2);
      end; }
    end;
  end;  // end while
  Flags.WriteBits(TLZ_MATCH, 1);
  inc(FlagsCount);
  EncodeLength(255);
  Flags.WriteBitsFlush;
  Literals.WriteBitsFlush;
  Pairs.WriteBitsFlush;
  Lengthes.WriteBitsFlush;
  Flags.Position:= 0;
  Literals.Position:= 0;
  Pairs.Position:= 0;
  Lengthes.Position:= 0;

  CompressedStreamWriteBits(Swap(Pairs.ReadBits(16)), 16);
  A:= 16;
  For I:= 0 to Flags.Size - 1 do
  begin
    Flag:= Flags.ReadBits(8);
    CompressedStreamWriteBits(Flag, 8);
    Mask:= $80;
    while Mask <> 0 do
    begin
      if (Mask and Flag) = 0 then
      begin
        dec(A, Lengthes.ReadBits(4));
        if A < 16 then
        begin
          CompressedStreamWriteBits(Swap(Pairs.ReadBits(16)), 16);
          inc(A, 16);
        end;
        dec(A, Lengthes.ReadBits(4));
        if A < 16 then
        begin
          CompressedStreamWriteBits(Swap(Pairs.ReadBits(16)), 16);
          inc(A, 16);
        end;
      end
      else
      begin
        CompressedStreamWriteBits(Literals.ReadBits(8), 8);
      end;
      Mask:= Mask shr 1;
      dec(FlagsCount);
      if FlagsCount = 0 then
      begin
        dec(CompressedSize, 2);
        exit;
      end;
    end;
  end;
end;  // Compress Loop

procedure TLZ.FindMatches(Offset, Len: PUINT);
var
  WPos: ULONG;
  TempLen: ULONG;
  MaxWPos: ULONG;
begin
  Len^:= TLZ_MINMATCHLEN - 1;
  if DataStreamPos < TLZ_WINLEN then
    WPos := 0
  else
    WPos:= DataStreamPos - TLZ_WINLEN;
  MaxWPos:= DataStreamPos;
  while (WPos < MaxWPos) do
    begin
      TempLen:= 0;
      while((Data[WPos + TempLen] = Data[DataStreamPos + TempLen]) and ((DataStreamPos + TempLen) < DataSize) and (TempLen < TLZ_MATCHLEN)) do
      begin
        inc(TempLen);
      end;  // end while
      if (TempLen >= Len^) then
      begin
        Len^:= TempLen;
        Offset^:= WPos;
      end;
      inc(WPos);
    end;
end;

function TLZ.Decompress(const Source; var Dest): ULONG;
procedure DeleteBits(Num: Byte);
var
  Temp: UINT;
  A: Byte;
begin
  BitStream[0]:= BitStream[0] shl Num;
  if BitCounts < Num then
  begin
    A:= Num - BitCounts;
    BitStream[0]:= BitStream[0] or (BitStream[1] shl A);
    CompressedStreamReadBits(@Temp, 8);
    BitStream[1]:= Temp;
    CompressedStreamReadBits(@Temp, 8);
    BitStream[1]:= BitStream[1] or (Temp shl 8);
    dec(Num, BitCounts);
    BitCounts:= 16;
  end;
  A:= BitCounts - Num;
  BitStream[0]:= BitStream[0] or (BitStream[1] shr A);
  BitStream[1]:= BitStream[1] and BitMask[A];
  Dec(BitCounts, Num);
end;
var
  Temp: UINT;
  Offset, Len: UINT;
  Flag, Mask: UINT;
  BitLength: Byte;
  t: UINT;
begin
  DataStreamPos:= 0;
  CompressedStreamPos:= 0;
  CompressedLong:= 0;
  CompressedBitsUsed:= 0;
  CompressedData:= @Source;
  Data:= @Dest;
  BitCounts:= 0;
  DeleteBits(16);
  while true do
  begin
    CompressedStreamReadBits(@Flag,8);
    Mask:= $80;
    while Mask <> 0 do
    begin
      if (Mask and Flag) <> 0 then
      begin
        CompressedStreamReadBits(@Temp, 8);
        Data[DataStreamPos]:= Temp;
        inc(DataStreamPos);
      end
      else
      begin
        Len:= DecodeLength(BitStream[0]);
        if Len = 255 then
        begin
          Data[DataStreamPos]:= CompressedStreamPos shr 8;
          Data[DataStreamPos+1]:= CompressedStreamPos;
          Result:= DataStreamPos;
          exit;
        end;
        inc(Len, 1);
        DeleteBits(BitsLen);
        Offset:= DecodePosition(BitStream[0]);
        DeleteBits(BitsLen);
        while Len > 0 do
        begin
          dec(Len);
          Data[DataStreamPos]:= Data[DataStreamPos - Offset - 1];
          inc(DataStreamPos);
        end;
      end;
      Mask:= Mask shr 1;
    end;
  end;
  Result:= DataStreamPos;
end;

end.
